import inspect
from abc import ABC, abstractmethod
from typing import Optional, Type, Union, Iterable, Tuple
from types import ModuleType
from importlib import import_module, reload as reload_module
import pkgutil

from ghostos_container import Provider, Container

__all__ = [
    'Modules', 'ImportWrapper', 'DefaultModules', 'DefaultModulesProvider',
]


class Modules(ABC):

    @abstractmethod
    def import_module(self, modulename: str) -> ModuleType:
        """
        引用一个模块或抛出异常.
        :param modulename: 模块全路径.
        :exception: ModuleNotFoundError
        """
        pass

    @abstractmethod
    def getsource(self, modulename: str) -> str:
        """
        get source from a module.
        """
        pass

    @abstractmethod
    def save_source(self, modulename: str, source: str, reload: bool = False) -> None:
        """
        save source code to a module, and reload it.
        """
        pass

    @abstractmethod
    def iter_modules(self, module: Union[str, ModuleType]) -> Iterable[Tuple[str, bool]]:
        """
        like pkgutil.iter_modules.
        :return: Iterable[(module_name, is_package)].
        """
        pass

    @abstractmethod
    def reload(self, module: Union[str, ModuleType]):
        pass


class ImportWrapper:
    def __init__(self, modules: Modules):
        self._modules = modules

    def __call__(self, modulename: str, globals_=None, locals_=None, from_list=None, level=-1):
        """
        可用于替换一个 module 上下文中的 __module__
        :param modulename: 完整的 module 路径.
        :param globals_: 默认不处理.
        :param locals_: 默认不处理.
        :param from_list: 应用的属性.
        :param level: 只能是 -1
        :return:
        """
        module = self._modules.import_module(modulename)
        result = []
        from_list = from_list if from_list is not None else []
        for name in from_list:
            if name in module.__dict__:
                result.append(module.__dict__[name])
            else:
                raise ModuleNotFoundError(f"Module {modulename} not found")
        return result


class DefaultModules(Modules):
    def import_module(self, modulename: str) -> ModuleType:
        return import_module(modulename)

    def getsource(self, modulename: str) -> str:
        module = self.import_module(modulename)
        source = inspect.getsource(module)
        return source

    def save_source(self, modulename: str, source: str, reload: bool = False) -> None:
        module = self.import_module(modulename)
        file = module.__file__
        with open(file, 'w') as f:
            f.write(source)
        if reload:
            self.reload(module)

    def iter_modules(self, module: Union[str, ModuleType]) -> Iterable[Tuple[str, bool]]:
        if isinstance(module, str):
            module_type = self.import_module(module)
        elif isinstance(module, ModuleType):
            module_type = module
        else:
            raise ValueError(f'Invalid module type: {type(module)}')
        prefix = module_type.__name__ + "."
        if not hasattr(module_type, "__path__"):
            return []
        path = module_type.__path__
        for i, name, is_pkg in pkgutil.iter_modules(path, prefix):
            yield name, is_pkg

    def reload(self, module: Union[str, ModuleType]):
        if isinstance(module, str):
            module = self.import_module(module)
        reload_module(module)


class DefaultModulesProvider(Provider[Modules]):
    def singleton(self) -> bool:
        return True

    def contract(self) -> Type:
        return Modules

    def factory(self, con: Container) -> Optional[Modules]:
        return DefaultModules()



import json
from typing import NamedTuple, Any, List, TypedDict, Optional
from types import ModuleType
from ghostos_moss.pycontext import PyContext
from pydantic import BaseModel, Field


def test_pycontext_join_baseline():
    left = PyContext()
    left.set_prop("foo", 123)
    right = PyContext()
    right.set_prop("bar", 234)
    joined = left.join(right)
    assert len(left.properties) == 1
    assert len(right.properties) == 1
    assert len(joined.properties) == 2


class Foo(BaseModel):
    foo: int = Field(default=123)


class Bar(TypedDict, total=False):
    bar: Optional[str]


def test_property_with_values():
    case = NamedTuple("Case", [("name", str), ("value", Any)])
    cases: List[case] = [
        case("foo", 123),
        case("bar", None),
        case("a", 1.0),
        case("", False),
        case("foo", Foo()),
        case("bar", Bar(bar="hello")),
    ]
    pycontext = PyContext()
    for c in cases:
        pycontext.set_prop(c.name, c.value)
        j = pycontext.model_dump_json()
        data = json.loads(j)
        new_one = PyContext(**data)
        value = new_one.get_prop(c.name)
        assert c.value == value, j


def test_property_with_local_module():
    module = ModuleType("__test__")
    code = """
from pydantic import BaseModel, Field
class Foo(BaseModel):
    foo: int = Field(default=123)
foo = Foo()
"""
    compiled = compile(code, "<test>", "exec")
    exec(compiled, module.__dict__)
    foo = module.__dict__["foo"]
    assert foo.foo == 123
    pycontext = PyContext()
    pycontext.set_prop(name="foo", value=foo)
    assert foo == pycontext.get_prop("foo", module)
    j = pycontext.model_dump(exclude_defaults=True)
    p = PyContext(**j)
    # 从当前 module 里重新还原出来.
    assert p.get_prop("foo", module) == foo



from typing import TypedDict
import inspect
from ghostos_moss import prompts
from ghostos_moss.prompts import reflect_locals_imported, reflect_code_prompt

from ghostos_moss.moss_impl import MossRuntimeImpl
from ghostos_moss.abcd import (
    MOSS_HIDDEN_MARK, MOSS_HIDDEN_UNMARK,
)


class Foo(TypedDict):
    foo: int


def test_reflect_locals_imported_baseline():
    assert inspect.ismodule(prompts)
    # inspect 也被 prompts 库引用了.
    assert not inspect.isbuiltin(inspect)
    attr_prompts = reflect_locals_imported("ghostos_moss.prompts", prompts.__dict__)
    data = {}
    array = []
    for name, prompt in attr_prompts:
        array.append((name, prompt))
        data[name] = prompt
    # 从 utils 模块里定义的.
    assert "get_callable_definition" in data
    # typing 库本身的不会出现.
    assert "Optional" not in data
    # 引用的抽象类应该存在.


def test_typed_dict_reflect_code():
    pr = reflect_code_prompt(Foo)
    source = inspect.getsource(Foo)
    assert len(source) > 0
    assert len(pr) > 0


def test_prompts_mark_judgement():
    parser = MossRuntimeImpl._parse_pycontext_code

    # test_no_hidden_markers
    code1 = """def foo():
        return 42
    print(foo())"""

    expected1 = code1
    assert parser(code1) == expected1

    # test_hidden_markers_with_model_visible
    code2 = f"""def foo():
        return 42
    {MOSS_HIDDEN_MARK}
    def hidden_function():
        return "hidden"
    {MOSS_HIDDEN_UNMARK}
    print(foo())"""

    expected2 = """def foo():
        return 42
    print(foo())"""
    assert parser(code2) == expected2

    # test_hidden_markers_with_model_invisible
    code3 = f"""def foo():
        return 42
    {MOSS_HIDDEN_MARK}
    def hidden_function():
        return "hidden"
    {MOSS_HIDDEN_UNMARK}
    print(foo())"""

    expected3 = f"""def foo():
        return 42
    {MOSS_HIDDEN_MARK}
    def hidden_function():
        return "hidden"
    {MOSS_HIDDEN_UNMARK}
    print(foo())"""
    assert parser(code3, exclude_hide_code=False) == expected3

    # test_multiple_hidden_sections
    code4 = f"""def foo():
        return 42
    {MOSS_HIDDEN_MARK}
    def hidden_function_1():
        return "hidden1"
    {MOSS_HIDDEN_UNMARK}
    def bar():
        return 24
    {MOSS_HIDDEN_MARK}
    def hidden_function_2():
        return "hidden2"
    {MOSS_HIDDEN_UNMARK}
    print(foo())"""

    expected4 = """def foo():
        return 42
    def bar():
        return 24
    print(foo())"""
    assert parser(code4) == expected4

    # test_nested_hidden_markers
    code5 = f"""def foo():
        return 42
    {MOSS_HIDDEN_MARK}
    def hidden_function_1():
        return "hidden1"
        {MOSS_HIDDEN_MARK}
        def hidden_function_2():
            return "hidden2"
        {MOSS_HIDDEN_UNMARK}
        return hidden_function_2()
    {MOSS_HIDDEN_UNMARK}
    print(foo())"""

    expected5 = """def foo():
        return 42
    print(foo())"""
    assert parser(code5) == expected5



from typing import NamedTuple, List
from ghostos_moss.utils import (
    get_class_def_from_source, replace_class_def_name, strip_source_indent, count_source_indent,
    parse_doc_string,
    escape_string_quotes,
)


def test_replace_class_def_name():
    Case = NamedTuple('Case', [('origin', str), ('name', str), ('expect', str)])

    cases: List[Case] = [
        Case(
            """
class Foo:
    class Bar:
        bar = 1
""",
            'Boo',
            """
class Boo:
    class Bar:
        bar = 1
"""
        ),
        Case(
            """
class Foo(ABC):
    class Bar:
        bar = 1
""",
            'Boo',
            """
class Boo(ABC):
    class Bar:
        bar = 1
"""
        )
    ]

    for c in cases:
        assert replace_class_def_name(c.origin.strip(), c.name) == c.expect.strip()


def test_strip_source_indent():
    Case = NamedTuple('Case', [('source', str), ('indent', int), ('expect', str)])
    cases: List[Case] = [
        Case(
            """
class Foo(ABC):
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
            0,
            """
class Foo(ABC):
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
        ),
        Case(
            """
        class Foo(ABC):
            \"""
            test
            \"""

            foo: int = 1

            def bar(self) -> str:
                return "bar"
""",
            8,
            """
class Foo(ABC):
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
        ),
        Case(
            """
    # if
    class Foo(ABC):
        \"""
        test
        \"""

        foo: int = 1

        def bar(self) -> str:
            return "bar"
""",
            4,
            """
# if
class Foo(ABC):
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
        ),
    ]
    for c in cases:
        assert count_source_indent(c.source.strip("\n")) == c.indent
        assert strip_source_indent(c.source.strip("\n")) == c.expect.strip()


def test_get_class_def_from_source():
    Case = NamedTuple('Case', [('source', str), ('expect', str)])
    cases: List[Case] = [
        Case(
            """
class Foo(ABC):
    \"""
    test
    \"""
    
    foo: int = 1
    
    def bar(self) -> str:
        return "bar"
""",
            """
class Foo(ABC):
"""),
        Case(
            """
class Foo(A, B, C, metaclass=E):
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
            """
class Foo(A, B, C, metaclass=E):
"""),
        Case(
            """
class Foo: # some
    \"""
    test
    \"""

    foo: int = 1

    def bar(self) -> str:
        return "bar"
""",
            """
class Foo: # some
"""),

    ]

    for c in cases:
        assert get_class_def_from_source(c.source) == c.expect.strip()


def test_escape_string_quotes():
    r = escape_string_quotes('hello """ \\""" world')
    assert r == 'hello \\""" \\""" world'


def test_parse_doc_string():
    Case = NamedTuple('Case', [('doc', str), ('expect', str), ('inline', bool)])
    cases: List[Case] = [
        Case(
            'hello world',
            '"""hello world"""',
            True,
        ),
        Case(
            'hello world',
            '"""\nhello world\n"""',
            False,
        ),
        Case(
            'hello """ world',
            '"""\nhello \\""" world\n"""',
            False,
        ),
    ]

    for c in cases:
        assert parse_doc_string(c.doc, inline=c.inline) == c.expect.strip()




from typing import Optional

from ghostos_moss.abcd import SelfUpdater
from ghostos_moss.modules import Modules
from ghostos_moss.pycontext import PyContext
from ghostos_common.helpers import get_attr_source_from_code
from ghostos_container import Container, Provider
import inspect
import os

__all__ = ["SelfUpdaterImpl", "SelfUpdaterProvider"]


class SelfUpdaterImpl(SelfUpdater):

    def __init__(self, pycontext: PyContext, modules: Modules):
        self.pycontext = pycontext
        self.modules = modules

    def append(self, code: str) -> None:
        source = self.getsource()
        new_code = "\n\n".join([source.rstrip(""), code.rstrip()]) + "\n"
        self.rewrite(new_code)

    def replace_attr(self, attr_name: str, code: str) -> None:
        source = self.getsource()
        origin = get_attr_source_from_code(source).get(attr_name, None)
        if origin:
            source = source.replace(origin, code, 1)
            self.rewrite(source)

    def getsource(self) -> str:
        if self.pycontext.code:
            return self.pycontext.code
        if self.pycontext.module:
            module = self.modules.import_module(self.pycontext.module)
            return inspect.getsource(module)
        return ""

    def rewrite(self, code: str) -> None:
        self.pycontext.code = code

    def save(self, reload: bool = False) -> None:
        if not self.pycontext.module:
            return None
        module = self.modules.import_module(self.pycontext.module)
        filename = module.__file__
        if os.path.exists(filename):
            with open(filename, "w") as f:
                f.write(self.pycontext.code)
            if reload:
                self.modules.reload(module)


class SelfUpdaterProvider(Provider[SelfUpdater]):

    def singleton(self) -> bool:
        return True

    def factory(self, con: Container) -> Optional[SelfUpdater]:
        pycontext = con.force_fetch(PyContext)
        modules = con.force_fetch(Modules)
        return SelfUpdaterImpl(pycontext, modules)



from abc import ABCMeta
from ghostos_moss import get_moss_compiler, MOSS_TYPE_NAME, Moss as Parent
import inspect


def test_module_without_moss_type_has_one():
    assert issubclass(type(Parent), ABCMeta)
    assert inspect.isabstract(Parent)
    compiler = get_moss_compiler()
    with compiler:
        rtm = compiler.compile(__name__)
        assert MOSS_TYPE_NAME in rtm.prompter().get_imported_attrs()

        prompt = rtm.prompter().dump_imported_prompt()
        assert "Moss(" in prompt


def test_with_default_moss_type():
    from ghostos_moss.examples.random_moss import RandomMoss
    compiler = get_moss_compiler()
    compiler.with_default_moss_type(RandomMoss)
    with compiler:
        rtm = compiler.compile(__name__)
        assert rtm.module().__dict__["Moss"] is RandomMoss
        assert rtm.moss_type() is RandomMoss
        prompt = rtm.prompter().dump_imported_prompt()
        assert "RandomMoss(" in prompt

        prompt = rtm.prompter().get_imported_attrs_prompt()
        assert "RandomMoss(" in prompt



from abc import abstractmethod
import inspect
from ghostos_moss.utils import *


def test_get_callable_definition():
    def foo() -> str:
        """
        test
        """
        return 'foo'

    assert get_callable_definition(foo) == """
def foo() -> str:
    \"""
    test
    \"""
    pass
""".strip()

    assert get_callable_definition(foo, alias="hello") == """
def hello() -> str:
    \"""
    test
    \"""
    pass
""".strip()

    assert get_callable_definition(foo, alias="hello", doc="") == """
def hello() -> str:
    pass
""".strip()

    assert get_callable_definition(foo, alias="hello", doc="world") == """
def hello() -> str:
    \"""
    world
    \"""
    pass
""".strip()


class Foo:

    def foo(self):
        return self

    @staticmethod
    def bar():
        return ""

    @classmethod
    def zoo(cls):
        return cls

    def cool(
            self,
            a: int,
            b: str,
    ) -> str:
        return ""

    @abstractmethod
    def do(self) -> str:  # test
        pass

    def doc(self) -> str:
        """
        test
        """
        return ""


def test_count_source_indent():
    import inspect
    source = inspect.getsource(Foo.foo)
    f = Foo()
    assert count_source_indent(source) == 4
    assert count_source_indent(inspect.getsource(Foo.bar)) == 4
    assert count_source_indent(inspect.getsource(f.zoo)) == 4


def test_strip_source_indent():
    f = Foo()
    assert strip_source_indent(inspect.getsource(f.foo)).startswith("def foo(")
    assert strip_source_indent(inspect.getsource(f.bar)).startswith("@staticmethod")
    assert strip_source_indent(inspect.getsource(f.zoo)).startswith("@classmethod")


def test_get_callable_definition_with_class():
    f = Foo()
    define = get_callable_definition(f.cool)
    expect = """
def cool(
        self,
        a: int,
        b: str,
) -> str:
    pass
"""
    assert define == expect.strip()

    # 包含 @staticmethod
    define = get_callable_definition(f.bar)
    expect = """
@staticmethod
def bar():
    pass
"""
    assert define == expect.strip()

    define = get_callable_definition(f.foo)
    expect = """
def foo(self):
    pass
"""
    assert define == expect.strip()

    # 删除了 @abstract
    define = get_callable_definition(Foo.do)
    expect = """
def do(self) -> str:  # test
    pass
"""
    assert define == expect.strip()

    define = get_callable_definition(Foo.do, "test")
    expect = """
def test(self) -> str:  # test
    pass
"""
    assert define == expect.strip()

    define = get_callable_definition(f.doc, "test")
    expect = '''
def test(self) -> str:
    """
    test
    """
    pass
'''
    assert define == expect.strip()

    define = get_callable_definition(f.doc, "test", doc="hello\nworld")
    expect = '''
def test(self) -> str:
    """
    hello
    world
    """
    pass
'''
    assert define == expect.strip()


def test_class_callable():
    class Bar:
        """
        good
        """

        def __call__(self, a: int, b: int) -> int:
            return a + b

    expect = """
def Bar(self, a: int, b: int) -> int:
    \"""
    good
    \"""
    pass
"""
    bar = Bar()
    assert get_callable_definition(bar) == expect.strip()


def test_get_callable_definition_with_abstractmethod():
    from abc import abstractmethod
    assert not inspect.isbuiltin(abstractmethod)
    code = get_callable_definition(abstractmethod)
    assert code != ""



from ghostos_moss import get_moss_compiler, PyContext
from ghostos_moss.examples import self_update_moss


def test_self_updater():
    compiler = get_moss_compiler()
    compiler.join_context(PyContext())
    runtime = compiler.compile(self_update_moss.__name__)
    with runtime:
        moss: self_update_moss.Moss = runtime.moss()
        assert moss.updater is not None
        moss.updater.append("foo = 123")
        assert "foo = 123" in runtime.dump_pycontext().code
        # moss.updater.save()



import time

from ghostos_moss import moss_container
from ghostos_moss.abcd import MossCompiler, Moss, MOSS_TYPE_NAME, MossRuntime
from ghostos_moss.pycontext import PyContext
from ghostos_moss.examples import baseline
from ghostos_moss.modules import ImportWrapper
from ghostos_container import Container


def test_baseline_exec():
    container = moss_container()
    compiler = container.force_fetch(MossCompiler)
    compiler = compiler.injects(bar=123)
    assert compiler is not None

    # join context
    compiler.join_context(PyContext(module=baseline.__name__))
    assert compiler.pycontext().module == baseline.__name__
    # 获取目标代码.
    code = compiler.pycontext_code()
    assert "from __future__" in code

    runtime = compiler.compile(None)
    assert runtime is not None

    module = runtime.module()
    # 名字相同.
    assert module.__name__ == baseline.__name__
    #
    importer = module.__dict__.get('__import__', None)
    assert isinstance(importer, ImportWrapper)

    # 先测试 ctx
    # with runtime.runtime_ctx():
    #     print("hello")
    # buffed = runtime.dump_std_output()
    # assert buffed.startswith("hello")

    exists_moss_type = module.__dict__.get(MOSS_TYPE_NAME)
    moss_type = runtime.moss_type()
    # 使用了默认的 MOSS
    assert issubclass(moss_type, Moss)
    assert moss_type is exists_moss_type

    moss = runtime.moss()
    assert isinstance(moss, Moss)
    # assert isinstance(moss, moss_type)

    prompter = runtime.prompter()
    assert prompter is not None
    prompt = prompter.dump_module_prompt()

    # plus 方法存在.
    assert 'def plus' in prompt
    # 在 moss 标记内的不展示.
    assert "__test__" not in prompt
    # 虽然import 了 inspect 的两个方法, 但一个的 prompt 被重置了.
    assert "def getmembers(" in prompt
    assert "def getsource(" not in prompt
    # 添加的意义不明的注释也应该存在了.
    assert "# hello world" in prompt

    # assert moss
    moss = runtime.moss()
    assert getattr(moss, "bar") is 123

    # 运行 main 方法.
    result = runtime.execute(target="test_main", local_args=["moss"])
    # main 方法的运行结果.
    assert result.returns == 3

    # 动态加载的 attr.
    assert "life" in result.pycontext.properties, f"life is not found in dumped pycontext {result.pycontext}"
    life = result.pycontext.properties["life"]
    assert life is not None
    # 生命周期被执行.
    value = result.pycontext.get_prop("life")
    assert isinstance(value, list)
    assert "__moss_compile__" in value, "in array test"
    assert '__moss_compile__' in value, "__moss_compile__ not found"
    assert 123 == result.pycontext.get_prop("bar")

    moss = runtime.moss()
    # 验证用 injections 注入.
    assert getattr(moss, 'bar') == 123
    # 验证依赖注入.
    foo = getattr(moss, 'foo')
    Foo = runtime.module().__dict__['Foo']
    assert Foo is baseline.Foo
    moss.fetch(Foo)
    assert foo is not None and isinstance(foo, Foo)
    assert foo.foo() == "hello"

    # 最后成功销毁.
    runtime.close()
    container.shutdown()


def test_baseline_in_test_mode():
    container = moss_container()
    compiler = container.force_fetch(MossCompiler)
    assert compiler is not None

    # join context
    compiler.join_context(PyContext(module=baseline.__name__))
    assert compiler.pycontext().module == baseline.__name__

    # 获取目标代码.
    with compiler:
        runtime = compiler.compile("__test__")
        assert runtime is not None

        module = runtime.module()
        # 名字相同.
        assert module.__name__ != baseline.__name__
        assert module.__name__ == "__test__"
        with runtime:
            moss = runtime.moss()
            moss.hello = "world"
            result = runtime.execute(target="test_main", local_args=["moss"])
            assert result.returns == 3
            assert result.pycontext.get_prop("hello") == "world"
    container.shutdown()


def test_baseline_with_pycontext_code():
    container = moss_container()
    compiler = container.force_fetch(MossCompiler)
    assert compiler is not None
    # join context
    line = "print('hello')"
    compiler.join_context(PyContext(module=baseline.__name__, code=line))
    assert line in compiler.pycontext_code()
    container.shutdown()


def test_moss_gc():
    from threading import Thread
    from gc import collect
    from ghostos_moss.moss_impl import MossStub, MossTempModuleType
    container = moss_container()
    container_count = Container.instance_count
    moss_stub_count = MossStub.instance_count

    def run(c: Container):
        compiler = c.force_fetch(MossCompiler)
        compiler = compiler.join_context(PyContext(module=baseline.__name__))
        runtime = compiler.compile("__test__")
        assert runtime.moss() is not None
        assert MossRuntime.instance_count > 0
        assert MossStub.instance_count > 0
        assert MossTempModuleType.__instance_count__ > 0
        with runtime:
            runtime.execute(target="test_main", local_args=["moss"])

    threads = []
    for i in range(10):
        t = Thread(target=run, args=(container,))
        t.start()
        threads.append(t)
    for t in threads:
        t.join()

    # assert gc success
    collect()
    time.sleep(0.05)
    assert MossTempModuleType.__instance_count__ < 10
    assert MossRuntime.instance_count == 0
    assert MossStub.instance_count <= moss_stub_count
    assert Container.instance_count < (10 + container_count)
